// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package schema is a legacy package that used to represent the SDK, which is now its own
// library external to OpenTofu Core https://github.com/hashicorp/terraform-plugin-sdk
// Some of it is still used by OpenTofu's remote state backends, but this entire
// package should be removed in the future.
package schema
