// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package repl provides the structs and functions necessary to run
// REPL for OpenTofu. The REPL allows experimentation of OpenTofu
// interpolations without having to run a OpenTofu configuration.
package repl
