// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package statefile deals with the file format used to serialize states for
// persistent storage and then deserialize them into memory again later.
package statefile
