// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package structured contains the structured representation of the JSON changes
// returned by the jsonplan package.
//
// Placing these in a dedicated package allows for greater reuse across the
// various type of renderers.
package structured
