// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package planproto is home to the Go stubs generated from the tfplan protobuf
// schema.
//
// This is an internal package to be used only by OpenTofu's planfile package.
// From elsewhere in OpenTofu, use the API exported by the planfile package
// itself.
package planproto
