// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package lang deals with the runtime aspects of OpenTofu's configuration
// language, with concerns such as expression evaluation. It is closely related
// to sibling package "configs", which is responsible for configuration
// parsing and static validation.
package lang
