// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package plans contains the types that are used to represent OpenTofu plans.
//
// A plan describes a set of changes that OpenTofu will make to update remote
// objects to match with changes to the configuration.
package plans
