// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package objchange deals with the business logic of taking a prior state
// value and a config value and producing a proposed new merged value, along
// with other related rules in this domain.
package objchange
