// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package checks contains the models for representing various kinds of
// declarative condition checks that can be defined in a OpenTofu module
// and then evaluated and reported by OpenTofu Core during plan and apply
// operations.
package checks
