// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package modsdir is an internal package containing the model types used to
// represent the manifest of modules in a local modules cache directory.
package modsdir
