// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package jsonchecks implements the common JSON representation of check
// results/statuses that we use across both the JSON plan and JSON state
// representations.
package jsonchecks
