// Copyright (c) The OpenTofu Authors
// SPDX-License-Identifier: MPL-2.0
// Copyright (c) 2023 HashiCorp, Inc.
// SPDX-License-Identifier: MPL-2.0

// Package provisioners contains the interface and primary types to implement a
// OpenTofu resource provisioner.
package provisioners
